#include <iostream>

// À compiler avec `g++ -std=c++23 exemple_classe_matrice.cpp` pour que operator[] avec plusieurs arguments soit légal

/*************/
/* matrice.h */
/*************/

class Matrice {
public:
	// Constructeurs
	Matrice ();
	Matrice (unsigned int cols, unsigned int rows, double valdef=0.);

	// Constructeur par copie
	Matrice (const Matrice&);

	// Constructeur par déplacement
	Matrice (Matrice&&);

	// Assignation par copie ou déplacement
	Matrice& operator= (Matrice);

	// Destructeur
	~Matrice();

	// Accès (setter/getter)
	double& operator[] (int row, int col);
	double operator[] (int row, int col) const;
	int nlignes () const { return rows; }
	int ncolonnes () const { return cols; }

private:
	double* data;
	unsigned int cols, rows;

	// calcule la position dans `data` de l'élément de matrice (row,col)
	size_t indice_data (int row, int col) const;
};

/************/
/* main.cpp */
/************/

Matrice test () {
	Matrice t = Matrice(2,2);
	t[0,0] = 1;
	return t;
}

int main () {

	{
		Matrice mat1 = Matrice(3, 5);
		mat1[1,1] = 4.2;

		std::cout << "----- Assignement -----\n";
		{
			Matrice mat2 = Matrice(1, 1);
			mat2 = mat1; // assignation par copie
			std::cout << "mat2[1,1] = " << mat2[1,1] << std::endl;
		}
		std::cout << "-----------------------\n";

		std::cout << "mat1[1,1] = " << mat1[1,1] << std::endl;

		std::cout << "----- Déplacement -----\n";
		{
			Matrice mat3 = std::move(mat1);
			std::cout << "mat3[1,1] = " << mat3[1,1] << std::endl;
			std::cout << "mat1.cols = " << mat1.ncolonnes() << std::endl;
		}
		std::cout << "-----------------------\n";
	}

	std::cout << "----- Fonction test -----\n";
	{
		Matrice mat4 = test();
		std::cout << "mat4[0,0] = " << mat4[0,0] << std::endl;
	}
	std::cout << "-----------------------\n";

	return 0;
}

/***************/
/* matrice.cpp */
/***************/

// Définition du contructeur de matrice vide
Matrice::Matrice(unsigned int cols_, unsigned int rows_, double valdef) 
	: cols(cols_), rows(rows_), data(nullptr)
{
	data = new double[cols*rows];
	for (unsigned int k = 0; k < cols*rows; k++)
		data[k] = valdef;
	std::cout << "Constructeur matrice " << cols << "x" << rows << " @" << data << ".\n";
}

// Définition du contructeur par défaut (matrice de taille nulle)
Matrice::Matrice() : Matrice(0,0) {}

// Définition du destructeur
Matrice::~Matrice() {
	if (data != nullptr)
		delete[] data;
	std::cout << "Destructeur @" << data << ".\n";
	data = nullptr;
}

// Définition du constructeur par copie
Matrice::Matrice (const Matrice& mat_à_copier)
	: cols(mat_à_copier.cols), rows(mat_à_copier.rows), data(nullptr)
{
	data = new double[cols*rows];
	std::cout << "Constructeur copie matrice " << cols << "x" << rows << " @" << data << ".\n";
	for (unsigned int k = 0; k < cols*rows; k++)
		data[k] = mat_à_copier.data[k];
}

// Définition du constructeur par déplacement
Matrice::Matrice (Matrice&& mat_à_déplacer)
	// on "vole" les attributs de `mat_à_déplacer`
	: cols(mat_à_déplacer.cols), rows(mat_à_déplacer.rows), data(mat_à_déplacer.data)
{
	std::cout << "Récupération de @" << mat_à_déplacer.data << ".\n";
	// on laisse `mat_à_déplacer` dans un état valide :
	mat_à_déplacer.cols = 0;
	mat_à_déplacer.rows = 0;
	mat_à_déplacer.data = nullptr;
}

// Définition de l'assignation par copie
Matrice& Matrice::operator= (Matrice temp) {
	                         // ^- copie
	std::swap(cols, temp.cols);
	std::swap(rows, temp.rows);
	std::swap(data, temp.data);
	return *this;
	// <- destruction de temp
}

// Accès/modification élément de matrice
double& Matrice::operator[] (int row, int col) {
	return data[ indice_data(row,col) ]; // renvoie une référence, modifiable
}

// Accès élément de matrice
double Matrice::operator[] (int row, int col) const {
	return data[ indice_data(row,col) ]; // renvoie par copie
}

// Indice d'un élément de matrice
size_t Matrice::indice_data (int row, int col) const {
	if (row < 0 or row >= rows)
		throw std::out_of_range("Matrice : numéro de ligne invalide");
	if (col < 0 or col >= cols)
		throw std::out_of_range("Matrice : numéro de colonne invalide");
	return row*cols + col;
}
